﻿/** 
 * Administration.cs
 * Eco version: Alpha 0.6.4.1
 * Mod version: 1.1.0
 * 
 * Author: Pradoxzon
 * https://pradoxzon.squarespace.com
 * 
 * 
 * This mod was made to provide server administrators with ways to dump 
 * information to a file.
 * 
 * Some ideas and code snippets are taken from ECO and SLG.
 * 
 * Mod created for: Pam
 * 
 * 
 * Available commands:
 *   autobackup [start|stop] (not implemented)
 *      Starts or stops auto backups.
 *      
 *   backup
 *      Performs a backup.
 *      
 *   buildtechtreefix (not implemented)
 *      Builds an image of the tech tree.
 *      
 *   exportchatlog
 *      Writes the chat log to a csv file.
 *      
 *   exportitemlist
 *      Exports a list of all items. Food items have their nutrient values listed.
 *      
 *   exportstats
 *      Exports stats for the server. This includes world stats and stats for players.
 *   
 *   save
 *      Saves the game.
 *      
 *   shutdown (not implemented)
 *      Closes the server. Saves the game before closing.
 *      
 */

namespace Eco.Pradoxzon.Administration
{
    using System;
    using System.Text;
    using System.IO;
    using System.Reflection;
    using System.Linq;
    using System.Collections.Generic;
    using System.Text.RegularExpressions;
    using Eco.Core;
    using Eco.Core.Plugins;
    using Eco.Core.Plugins.Interfaces;
    using Eco.Gameplay.Systems.Chat;
    using Eco.Gameplay.Players;
    using Eco.Gameplay.Items;
    using Eco.Gameplay.Stats;
    using Eco.Shared.Services;
    using Eco.Shared.Utils;
    using Eco.Simulation.Time;
    using Eco.Stats;
    using LiteDB;


    // Class that contains all of the chat commands provided in this mod
    class AdminCommands : IChatCommandHandler
    {
        private const string Unfinished = "[Administration] This command is not yet implemented. " 
            + "Contact Pam or Pradoxzon on the Eco Discord server for more information.";

        // Output directory for files
        private static string OutDirectory = String.Format("{0}{1}ExportedDocs", 
            new FileInfo(Assembly.GetEntryAssembly().Location).Directory.ToString(), Path.DirectorySeparatorChar);

        // Check that the directory exists
        private static void CheckDir()
        {
            if (!Directory.Exists(OutDirectory))
            {
                Directory.CreateDirectory(OutDirectory);
            }
        }


        /**
         * Command: /autobackup [start|stop]
         * See file header for a detailed description.
         */
        //[ChatCommand("autobackup", "Starts or stops auto backups. Usage: /autobackup [start|stop]", ChatAuthorizationLevel.Admin)]
        public static void AutoBackup(User user, string startStop = "start")
        {
            //Dictionary<string, Action> action = new Dictionary<string, Action>();
            //BackupPlugin.GetCommands(action);
            if (!(startStop == "start" || startStop == "stop"))
            {
                Core.Echo("[Administration] Usage: /autobackup [start|stop]", user);
                return;
            }
            else if (startStop == "stop")
            {

            }
            else
            {

            }

            // TODO
            Core.Echo(Unfinished, user);
        }


        /**
         * Command: /backup
         * See file header for a detailed description.
         */
        [ChatCommand("backup", "Performs a backup.", ChatAuthorizationLevel.Admin)]
        public static void Backup(User user)
        {
            StorageManager.Backup();

            Core.Echo("[Administration] Backed up the game.", user);
        }


        /**
         * Command: /buildtechtreefixed
         * See file header for a detailed description.
         */
        //[ChatCommand("buildtechtreefixed", "Builds an image of the tech tree.", ChatAuthorizationLevel.Admin)]
        public static void BuildTechTree(User user)
        {
            CheckDir();

            // TODO
            Core.Echo(Unfinished, user);
        }


        /**
         * Command: /exportchatlog
         * See file header for a detailed description.
         */
        [ChatCommand("exportchatlog", "Dumps the chat log to a csv file.", ChatAuthorizationLevel.Admin)]
        public static void ExportChatLog(User user)
        {
            CheckDir();
            Regex htmltags = new Regex("<[^>].+?>");

            // StringBuilder to output all chat to
            StringBuilder outString = new StringBuilder();

            // First line
            outString.AppendLine("Tag,TimeSent,Sender,Text");

            // Every other line
            //ChatServer.GetPlayerMessages().Select(chat => outString.AppendLine(String.Format("{0},{1},{2},{3}",
            //        chat.Tag, chat.TimeSeconds.ToString(), chat.Sender, htmltags.Replace(chat.Text, string.Empty).Replace("\n", string.Empty))));
            foreach (var chat in ChatServer.GetPlayerMessages())
            {
                outString.AppendLine(String.Format("{0},{1},{2},{3}",
                    chat.Tag, chat.TimeSeconds.ToString(), chat.Sender, htmltags.Replace(chat.Text, string.Empty).Replace("\n", string.Empty)));
            }

            // Save to file in ExportedDocs/ChatLogs
            if (!Directory.Exists(Path.Combine(OutDirectory, "ChatLogs")))
            {
                Directory.CreateDirectory(Path.Combine(OutDirectory, "ChatLogs"));
            }
            string fileName = String.Format("ChatLogs{0}ChatLog_{1}.csv",
                Path.DirectorySeparatorChar, DateTime.Now.ToString("yyMMddhhmmss"));
            string dir = Path.Combine(OutDirectory, fileName);
            StreamWriter output = new StreamWriter(dir);
            output.Write(outString.ToString());
            output.Close();

            // Announce that the command finished
            Core.Echo(String.Format("[Administration] Dumped {0}", fileName), user);
        }


        /**
         * Command: /exportitemlist
         * See file header for a detailed description.
         */
        [ChatCommand("exportitemlist", "Exports a list of all items.", ChatAuthorizationLevel.Admin)]
         public static void ExportItemList(User user)
        {
            CheckDir();

            // Store all the items in lists based on their category
            Dictionary<string, SortedDictionary<string, string>> masterList = new Dictionary<string, SortedDictionary<string, string>>()
            {
                { "Actionbar Items", new SortedDictionary<string, string>() },
                { "AvatarPart", new SortedDictionary<string, string>() },
                { "Block Items", new SortedDictionary<string, string>() },
                { "Food", new SortedDictionary<string, string>() },
                { "Items", new SortedDictionary<string, string>() },
                { "Modules", new SortedDictionary<string, string>() },
                { "Road Items", new SortedDictionary<string, string>() },
                { "Skill Books", new SortedDictionary<string, string>() },
                { "Skill Scrolls", new SortedDictionary<string, string>() },
                { "Skills", new SortedDictionary<string, string>() },
                { "Tools", new SortedDictionary<string, string>() },
                { "World Object Items", new SortedDictionary<string, string>() }
            };

            // Iterate over all items
            foreach (Item item in Item.AllItems)
            {
                // Add the item to the apropriate list
                if (!masterList.ContainsKey(item.Group))
                {
                    masterList.Add(item.Group, new SortedDictionary<string, string>());
                }
                string iType = item.Type.ToString().Substring(item.Type.ToString().LastIndexOf('.') + 1);
                if (masterList[item.Group].ContainsKey(iType))
                {
                    continue;
                }
                masterList[item.Group].Add(iType, iType);

                // Get the nutrients
                if (item is FoodItem)
                {
                    var foodItem = item as FoodItem;
                    StringBuilder append = new StringBuilder();

                    append.AppendLine(String.Format("    calories: {0}", foodItem.Calories.ToString("0.0")));
                    append.AppendLine(String.Format("    carbs: {0}", foodItem.Nutrition.Carbs.ToString("0.0")));
                    append.AppendLine(String.Format("    protein: {0}", foodItem.Nutrition.Protein.ToString("0.0")));
                    append.AppendLine(String.Format("    fat: {0}", foodItem.Nutrition.Fat.ToString("0.0")));
                    append.Append(String.Format("    vitamins: {0}", foodItem.Nutrition.Vitamins.ToString("0.0")));

                    masterList[item.Group][iType] += "\n" + append.ToString();
                }
                //// Get housing values
                //else if (item is WorldObjectItem)
                //{

                //}
            }

            // Save to files in ExportedDocs/ItemDump
            if (!Directory.Exists(String.Format("{0}{1}ItemDump", OutDirectory, Path.DirectorySeparatorChar)))
            {
                Directory.CreateDirectory(String.Format("{0}{1}ItemDump", OutDirectory, Path.DirectorySeparatorChar));
            }
            string partialFileName = String.Format("{0}{1}ItemDump{2}", 
                OutDirectory, Path.DirectorySeparatorChar, Path.DirectorySeparatorChar);
            foreach (string category in masterList.Keys)
            {
                StreamWriter output = new StreamWriter(String.Format("{0}{1}.txt", partialFileName, category));

                foreach (string item in masterList[category].Keys)
                {
                    output.WriteLine(masterList[category][item]);
                }

                output.Close();
            }
            

            // Announce that the command finished
            Core.Echo(String.Format("[Administration] Dumped files to ExportedDocs{0}ItemDump", 
                Path.DirectorySeparatorChar), user);
        }


        /**
         * Command: /exportstats
         * See file header for a detailed description.
         */
        [ChatCommand("exportstats", "Exports stats of the players.", ChatAuthorizationLevel.Admin)]
        public static void ExportStats(User user)
        {
            CheckDir();
            string partDir = String.Format("ExportedStats-{0}", DateTime.Now.ToString("yyyy-MM-ddTHH-mm"));
            string directory = Path.Combine(OutDirectory, partDir);
            Directory.CreateDirectory(directory);

            // Export World stats (using same methods as GameplayStatsPlugin.ExportWorldStats())
            Dictionary<int, Dictionary<string, float>> secondsToValues = new Dictionary<int, Dictionary<string, float>>();

            foreach (StatManager statManager in WorldStats.Managers.Concat(OrganismStats.GetPopulationStats()))
            {
                IList<FloatStatEntry> entries = statManager.GetGraphableData(WorldTime.Seconds);

                foreach (FloatStatEntry entry in entries)
                {
                    Dictionary<string, float> values = secondsToValues.GetOrAdd(entry.TimeSeconds);
                    values[String.Format("{0} ({1})", statManager.DisplayName, statManager.Unit)] = entry.Value;
                }
            }

            List<string> worldHeaders = new List<string>() { "Seconds since server start:" };
            Dictionary<string, List<string>> rowData = new Dictionary<string, List<string>>();

            foreach (KeyValuePair<int, Dictionary<string, float>> kvpTimeData in secondsToValues.OrderBy(kvp => kvp.Key))
            {
                worldHeaders.Add(Text.Int(kvpTimeData.Key));
                foreach (KeyValuePair<string, float> data in kvpTimeData.Value)
                {
                    List<string> entry = rowData.GetOrAdd(data.Key, () =>
                    {
                        var e = new List<string>();
                        e.Add(data.Key);
                        return e;
                    });

                    entry.Add(Text.Num(data.Value));
                }
            }

            IEnumerable<List<string>> rows = new List<List<string>>() { worldHeaders }.Concat(rowData.Values);
            string worldFileText = string.Join("\n", rows.Select(line => string.Join(",", line.Select(entry => String.Format("\"{0}\"", entry)))));

            using (StreamWriter file = new StreamWriter(Path.Combine(directory, "WorldStats.csv")))
                file.Write(worldFileText);


            // Export Player stats (using same methods as GameplayStatsPlugin.ExportPlayerActions())
            foreach (IPlayerStatManager statManager in PlayerActions.StatManagers)
            {
                List<string> playerHeaders = null;
                var entries = new List<List<string>>();

                foreach (IPlayerAction action in statManager.GetAllActions())
                {
                    BsonDocument document = BsonMapper.Global.ToDocument(action.GetType(), action);

                    if (playerHeaders == null)
                    {
                        playerHeaders = document.Keys.Where(str => str != "_id").ToList(); // _id is a special field for LiteDB
                        entries.Add(playerHeaders);
                    }

                    List<string> entry = new List<string>();
                    entries.Add(entry);

                    foreach (string header in playerHeaders)
                    {
                        if (document.ContainsKey(header))
                            entry.Add(Regex.Replace(document[header].ToString().Replace(',', ' '), "\\\\?\"", string.Empty));
                        else
                            entry.Add(string.Empty);
                    }
                }

                if (entries.Any())
                {
                    string playerFileText = string.Join("\n", entries.Select(line => string.Join(",", line.Select(entry => String.Format("\"{0}\"", entry)))));
                    using (StreamWriter file = new StreamWriter(Path.Combine(directory, 
                        String.Format("PlayerAction-{0} ({1}).csv", statManager.DisplayName, statManager.Unit))))
                    {
                        file.Write(playerFileText);
                    }
                }
            }


            // Indicate to the player that the command is done
            Core.Echo(String.Format("[Administration] Exported stats to {0}", Path.Combine("ExportedDocs", partDir)), user);
        }


        /**
         * Command: /save
         * See file header for a detailed description.
         */
        [ChatCommand("save", "Saves the state of the game.", ChatAuthorizationLevel.Admin)]
        public static void Save(User user)
        {
            PluginManager.Controller.ForEach(plugin =>
            {
                if (plugin is IConfigurablePlugin)
                {
                    (plugin as IConfigurablePlugin).SaveConfig();
                }
            });

            Core.Echo("[Administration] Saved the game.", user);
        }


        /**
         * Command: /shutdown
         * See file header for a detailed description.
         */
        //[ChatCommand("shutdown", "Stops the server.", ChatAuthorizationLevel.Admin)]
        public static void Shutdown(User user)
        {
            // Save the game before shutting down
            Save(user);

            // TODO
            Core.Echo(Unfinished, user);
        }


        private static string FilenameFix(string name)
        {
            return name.Replace('\\', '-');
        }
    }




    // Class that contains functions used by commands
    class Core
    {
        /**
         *  Sends a chat notification to a user.
         *  
         *  \param message The message to send to the player
         *  \param user The user who will recieve the message
         */
        public static void Echo(string message, User user)
        {
            ChatManager.ServerMessageToPlayer(string.Format(message), user, false, DefaultChatTags.Notifications, ChatCategory.Info);
        }
    }
}
